@@ -1,5 +1,5 @@
 /*  dctrl-tools - Debian control file inspection tools
-    Copyright © 2005, 2006, 2007, 2008 Antti-Juhani Kaijanaho
+    Copyright © 2005, 2006, 2007, 2008, 2009, 2010 Antti-Juhani Kaijanaho
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,9 +11,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <argp.h>
@@ -98,6 +98,13 @@ size_t linewrap(char **res, char const *orig, size_t orig_len,
 
 	mblen(NULL, 0);
 	for (size_t i = 0; i < orig_len; /**/) {
+                if (orig[i] == '\n') {
+                        i++;
+                        ll = 0;
+                        INSERT('\n');
+                        num_lines++;
+                        continue;
+                }
 		if (ll == col_width) {
 			if (rv[bpr] == '\n') {
 				// no suitable breakpoint on this line,
@@ -335,10 +342,16 @@ static struct argp argp = { options, parse_opt, 0, progdoc };
 static size_t mbs_len(char const *mbs, size_t n, char const *fname)
 {
 	if (n == (size_t)(-1)) n = strlen(mbs);
+        size_t mlen = 0;
 	size_t len = 0;
 	mblen(NULL, 0);
 	for (size_t k = 0; k < n;/**/) {
-		if (mbs[k] == '\n') break;
+		if (mbs[k] == '\n') {
+                        if (len > mlen) mlen = len;
+                        len = 0;
+                        k++;
+                        continue;
+                }
 		len++;
 		int delta = mblen(mbs + k, n - k);
 		if (delta <= 0) {
@@ -349,7 +362,7 @@ static size_t mbs_len(char const *mbs, size_t n, char const *fname)
 			k += delta;
 		}
 	}
-	return len;
+	return len > mlen ? len : mlen;
 }
 
 int main(int argc, char * argv[])
@@ -372,23 +385,40 @@ int main(int argc, char * argv[])
 		};
 	}
 
-	if (args.need_preprocessing) {
+	if (args.num_columns == 0 || args.need_preprocessing) {
 		struct para_bundle pb;
 
 		bundle_init(&pb);
 		
 		for (size_t i = 0; i < args.num_fnames; i++) {
-			bundle_slurp(&pb, args.fname[i]);
+			bundle_slurp(&pb, args.fname[i],
+                                     args.num_columns == 0);
 		}
 
-		size_t n = fieldtrie_count();
+		const size_t n = fieldtrie_count();
+
+                if (args.num_columns == 0) {
+                        size_t m = n <= MAX_COLUMNS ? n : MAX_COLUMNS;
+                        for (size_t i = 0; i < m; i++) {
+                                struct field_attr *fa = fieldtrie_get(i);
+                                args.columns[i].heading =
+                                        strndup(fa->name, fa->namelen);
+                                if (args.columns[i].heading == NULL)
+                                        fatal_enomem(NULL);
+                                args.columns[i].field_inx = fa->inx;
+                                args.columns[i].column_width = -1;
+                        }
+                        args.num_columns = m;
+                }
+
 		size_t lens[n];
 		for (size_t i = 0; i < n; i++) lens[i] = 0;
 
 		for (size_t i = 0; i < pb.num_paras; i++) {
 			struct paragraph *p = pb.paras[i];
-			assert(p->nfields == n);
+			assert(p->nfields <= n);
 			for (size_t j = 0; j < p->nfields; j++) {
+                                if (!p->fields[j].present) continue;
 				struct fsaf_read_rv r = get_field(p, j, -1);
 				size_t len = mbs_len(r.b, r.len,
 						     p->common->fp->fname);
